import typing as t
from typing_extensions import Unpack

if t.TYPE_CHECKING:  # pragma: no cover
    from jsonschema import Draft7Validator


class RefKwargs(t.TypedDict, total=False):
    input_message_ref: str
    output_ref: str
    output_message_ref: str


class Schema:
    """
    Helper for validating request, result, and serialized result schemas
    """

    def __init__(
        self,
        schema: t.Any,
        input_message_ref: str = "#/definitions/AnyInputMessage",
        output_ref: str = "#/definitions/Output",
        output_message_ref: str = "#/definitions/AnyOutputMessage",
    ):
        from jsonschema import RefResolver

        self.resolver = RefResolver.from_schema(schema)
        self.input_message = (
            None if input_message_ref is None else self.validator_for(input_message_ref)
        )
        self.output = None if output_ref is None else self.validator_for(output_ref)
        self.output_message = (
            None
            if output_message_ref is None
            else self.validator_for(output_message_ref)
        )

    @classmethod
    def load_from_path(
        cls, schema_filename: str, **kwargs: Unpack[RefKwargs]
    ) -> "Schema":
        from missouri import json

        return cls(schema=json.load(schema_filename), **kwargs)

    @classmethod
    def load_relative_to_file(
        cls, filename: str, path_components: list[str], **kwargs: Unpack[RefKwargs]
    ) -> "Schema":
        """
        By convention, the schema is placed in
        `types/src/generated/schema.json` relative to the handler.
        """
        import os

        return cls.load_from_path(
            schema_filename=os.path.join(
                os.path.dirname(filename),
                *path_components,
            ),
            **kwargs,
        )

    def validator_for(self, ref: str) -> "Draft7Validator":
        from jsonschema import Draft7Validator

        return Draft7Validator({"$ref": ref}, resolver=self.resolver)



import os
import typing as t

from .common.cli_common import InferredFunctionEnvironment


def local_path_for_built_lambda(function_name: str) -> str:
    return os.path.join("lambdas", f"{function_name}.zip")


def local_path_for_manifest(function_name: str) -> str:
    return os.path.join("lambdas", f"{function_name}.manifest.json")


def _get_s3_key_base(
    function_name: str,
    version: t.Optional[str],
    sha1: t.Optional[str],
    with_manifest: bool,
) -> str:
    if version:
        return f"{function_name}/tags/{version}"
    else:
        return f"{function_name}/branches/{sha1}"


def s3_key_for_built_lambda(
    function_name: str,
    version: t.Optional[str] = None,
    sha1: t.Optional[str] = None,
    with_manifest: bool = False,
) -> str:
    return f"{_get_s3_key_base(function_name, version, sha1, with_manifest)}.zip"


def s3_key_for_manifest(
    function_name: str,
    version: t.Optional[str] = None,
    sha1: t.Optional[str] = None,
    with_manifest: bool = False,
) -> str:
    return (
        f"{_get_s3_key_base(function_name, version, sha1, with_manifest)}.manifest.json"
    )


def publish_file_to_s3(
    key: str, filename: str, bucket_name: str, verbose: bool
) -> None:
    import boto3

    client = boto3.client("s3")

    def exists() -> bool:
        try:
            client.head_object(Bucket=bucket_name, Key=key)
        except client.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == 404:
                return True
        return False

    if exists():
        raise ValueError(f"s3://{bucket_name}/{key} already exists on S3")

    client.upload_file(
        Filename=filename,
        Bucket=bucket_name,
        Key=key,
    )


class PublishCommonKwargs(t.TypedDict):
    bucket_name: str
    verbose: bool


def perform_publish(
    environment: InferredFunctionEnvironment,
    verbose: bool,
    bucket_name: str,
    with_manifest: bool,
) -> None:
    function_name = environment.function_name
    version = environment.version
    sha1 = environment.sha1

    assert (version is not None) != (sha1 is not None)

    common_kwargs: PublishCommonKwargs = dict(
        bucket_name=bucket_name,
        verbose=verbose,
    )

    publish_file_to_s3(
        key=s3_key_for_built_lambda(
            function_name=function_name,
            version=version,
            sha1=sha1,
            with_manifest=with_manifest,
        ),
        filename=local_path_for_built_lambda(function_name=function_name),
        **common_kwargs,
    )

    if with_manifest:
        publish_file_to_s3(
            key=s3_key_for_manifest(
                function_name=function_name,
                version=version,
                sha1=sha1,
                with_manifest=with_manifest,
            ),
            filename=local_path_for_manifest(function_name=function_name),
            **common_kwargs,
        )



#!/usr/bin/env python3
import os


def set_path() -> None:
    import sys

    sys.path.insert(
        0, os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    )


set_path()

import click  # noqa: E402

from werkit.common.cli_common import (  # noqa: E402
    infer_commit_from_ci_environment,
    infer_function_environment_from_ci_environment,
)
from werkit.publish_to_s3 import (  # noqa: E402
    perform_publish,
)


@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option("-v", "--verbose", is_flag=True, default=True, help="Be extra talkative")
@click.option("-b", "--bucket-name", help="Archive bucket name", required=True)
@click.option(
    "--with-manifest/--without-manifest",
    help="Function inclues a manifest",
    default=False,
    is_flag=True,
)
def infer_and_publish(verbose: bool, bucket_name: str, with_manifest: bool) -> None:
    environment = infer_function_environment_from_ci_environment()
    perform_publish(
        environment=environment,
        verbose=verbose,
        bucket_name=bucket_name,
        with_manifest=with_manifest,
    )


@cli.command()
@click.option("-v", "--verbose", is_flag=True, default=True, help="Be extra talkative")
@click.option("-f", "--function-name", help="Function name", required=True)
@click.option("-b", "--bucket-name", help="Archive bucket name", required=True)
@click.option(
    "--with-manifest/--without-manifest",
    help="Function inclues a manifest",
    default=False,
    is_flag=True,
)
def publish_using_sha1(
    verbose: bool, function_name: str, bucket_name: str, with_manifest: bool
) -> None:
    environment = infer_commit_from_ci_environment(function_name)
    perform_publish(
        environment=environment,
        verbose=verbose,
        bucket_name=bucket_name,
        with_manifest=with_manifest,
    )


if __name__ == "__main__":
    cli()



import typing as t

from werkit.compute import Destination

if t.TYPE_CHECKING:
    from mypy_boto3_lambda.client import LambdaClient

_lambda_client = None


def get_lambda_client() -> "LambdaClient":
    import boto3

    global _lambda_client
    if not _lambda_client:  # pragma: no cover
        _lambda_client = boto3.client("lambda")

    return _lambda_client


class LambdaDestination(Destination):
    def __init__(
        self,
        task_identifier: t.Optional[str] = None,
        function_name: t.Optional[str] = None,
        qualifier: t.Optional[str] = None,
    ):
        self.task_identifier = task_identifier
        self.function_name = function_name
        self.qualifier = qualifier

    def send(self, message_key: t.Any, output_message: t.Any) -> None:
        from missouri import json
        import os

        task_identifier = self.task_identifier or os.environ.get(
            "DESTINATION_LAMBDA_TASK_IDENTIFIER", None
        )
        if task_identifier is None:
            raise ValueError(
                "Either DESTINATION_LAMBDA_TASK_IDENTIFIER must be set or task_identifier must be provided"
            )

        function_name = self.function_name or os.environ.get(
            "DESTINATION_LAMBDA_FUNCTION_NAME", None
        )
        if function_name is None:
            raise ValueError(
                "Either DESTINATION_LAMBDA_FUNCTION_NAME must be set or function_name must be provided"
            )

        qualifier = self.qualifier or os.environ.get(
            "DESTINATION_LAMBDA_QUALIFIER", None
        )

        # Keep this in sync with the DestinationMessage interface.
        message = {"taskIdentifier": self.task_identifier, "payload": output_message}

        # TODO: Dry this while satisfying the type checker...
        if qualifier is None:
            get_lambda_client().invoke(
                FunctionName=function_name,
                InvocationType="Event",
                Payload=json.dumps(message),
            )
        else:
            get_lambda_client().invoke(
                FunctionName=function_name,
                InvocationType="Event",
                Payload=json.dumps(message),
                Qualifier=qualifier,
            )



import asyncio
import concurrent
import datetime
import os
import typing as t
from botocore.exceptions import ClientError

from werkit.compute import Manager
from .parallel import parallel_map_on_lambda
from .schema import SCHEMA
from ...compute._serialization import serialize_exception

LAMBDA_WORKER_FUNCTION_NAME = "LAMBDA_WORKER_FUNCTION_NAME"
env_worker_lambda_function_name = os.environ.get(LAMBDA_WORKER_FUNCTION_NAME)

LAMBDA_WORKER_TIMEOUT = "LAMBDA_WORKER_TIMEOUT"
env_worker_lambda_timeout = (
    int(os.environ[LAMBDA_WORKER_TIMEOUT])
    if LAMBDA_WORKER_TIMEOUT in os.environ
    else None
)


def transform_result(
    message_key: t.Any, result: t.Any, start_time: datetime.datetime
) -> dict[str, t.Any]:
    if (
        isinstance(result, ClientError)
        or isinstance(result, asyncio.TimeoutError)
        # TODO: Could this be replaced with just the following clause?
        or isinstance(result, Exception)
    ):
        return serialize_exception(
            message_key=message_key,
            exception=result,
            error_origin="orchestration",
            start_time=start_time,
        )
    elif isinstance(result, dict) and "errorMessage" in result:
        # https://docs.aws.amazon.com/lambda/latest/dg/python-exceptions.html
        # Unhandled exception in Lambda, with `errorMessage`, `errorType`, and
        # `stackTrace`.
        return {
            "message_key": message_key,
            "success": False,
            "result": None,
            "error": result["stackTrace"]
            + [f"{result['errorType']}: {result['errorMessage']}"],
            "error_origin": "system",
            "start_time": start_time.astimezone().isoformat(),
            "duration_seconds": -1,
        }
    else:
        return result


# TODO: This handler should have a unit test which uses a stubbed lambda. This
# would dramatically simplify debugging this code.
def handler(
    event: dict[str, t.Any],
    context: dict[str, t.Any],
    worker_lambda_function_name: t.Optional[str] = env_worker_lambda_function_name,
    timeout: int = env_worker_lambda_timeout or 120,
) -> dict[str, t.Any]:
    print("input_message", event)

    manager = Manager(input_message=event, schema=SCHEMA)

    def work(_: t.Any) -> dict[str, t.Any]:
        if not worker_lambda_function_name:
            raise Exception(
                f"Environment variable {LAMBDA_WORKER_FUNCTION_NAME} must be defined, "
                + "or default kwArg lambda_worker_function_name must be bound to the handler"
            )
        item_collection = event["itemCollection"]
        event_loop = asyncio.get_event_loop()
        with concurrent.futures.ThreadPoolExecutor(max_workers=1023) as executor:
            all_results = event_loop.run_until_complete(
                parallel_map_on_lambda(
                    message_key=manager.message_key,
                    item_property_name=event["itemPropertyName"],
                    item_collection=item_collection,
                    common_input=event["commonInput"],
                    worker_lambda_function_name=worker_lambda_function_name,
                    timeout=timeout,
                    event_loop=event_loop,
                    executor=executor,
                )
            )
            return dict(
                zip(
                    item_collection.keys(),
                    [
                        transform_result(
                            message_key=manager.message_key,
                            result=result,
                            start_time=manager.start_time,
                        )
                        for result in all_results
                    ],
                )
            )

    return manager.work(work, should_send=False, should_return=True)



import typing as t
from ._dependency_graph import ComputeNode, DependencyGraph


class Store:
    def __init__(self, dependency_graph: DependencyGraph):
        self.dependency_graph = dependency_graph
        self.store: dict[str, t.Any] = {}

    def _assert_known_keys(self, keys: t.Iterable[str]) -> None:
        unknown_keys = set(keys) - set(self.dependency_graph.keys())
        if len(unknown_keys):
            raise KeyError(
                f"Unknown {'key' if len(unknown_keys) == 1 else 'keys'}: {', '.join(sorted(list(unknown_keys)))}"
            )

    def deserialize(self, **kwargs: t.Any) -> None:
        self._assert_known_keys(kwargs.keys())
        deserialized = {
            name: self.dependency_graph.all_nodes[name].deserialize(value)
            for name, value in kwargs.items()
        }
        self.store.update(deserialized)

    def normalize(self, **kwargs: t.Any) -> dict[str, t.Any]:
        return {
            name: self.dependency_graph.all_nodes[name].normalize(
                name=name, value=value
            )
            for name, value in kwargs.items()
        }

    def set(self, **kwargs: t.Any) -> None:
        self._assert_known_keys(kwargs.keys())
        normalized = self.normalize(**kwargs)
        self.store.update(normalized)

    def serialize(self, targets: t.Optional[t.List[str]] = None) -> dict[str, t.Any]:
        if targets is not None:
            self._assert_known_keys(targets)
            missing_keys = set(targets) - set(self.store.keys())
            if missing_keys:
                preamble = f"{'Key has' if len(missing_keys) == 1 else 'Keys have'} not been evaluated"
                raise ValueError(f"{preamble}: {', '.join(sorted(list(missing_keys)))}")
        return {
            name: self.dependency_graph.all_nodes[name].serialize_value(value)
            for name, value in self.store.items()
            if targets is None or name in targets
        }


class StateManager(Store):
    def __init__(self, instance: t.Any):
        super().__init__(
            dependency_graph=DependencyGraph.from_class(instance.__class__)
        )
        self.instance = instance

    def evaluate(
        self, targets: t.Optional[t.List[str]] = None, handle_exceptions: bool = False
    ) -> None:
        import functools
        from artifax import Artifax

        if targets is not None:
            if len(targets) == 0:
                # TODO: afx.build(targets=[]) raises an exception. This seems
                # like a bug in Artifax.
                return
            else:
                self._assert_known_keys(targets)

        def wrap_node(name: str, node: ComputeNode) -> t.Callable:
            wrapped = node.bind(self.instance)

            def wrapper(*args: list[t.Any]) -> t.Any:
                value = wrapped(*args)
                return node.normalize(name, value)

            functools.update_wrapper(wrapper, wrapped)
            return wrapper

        afx = Artifax(
            {
                name: wrap_node(name, node)
                for name, node in self.dependency_graph.compute_nodes.items()
            }
        )
        if self.store:
            afx.set(**self.store)

        if targets is None:
            self.store.update(**afx.build(targets=targets))
        else:
            values = afx.build(targets=targets)
            self.store.update(**dict(zip(targets, values)))

    def get(self, name: str) -> t.Any:
        self._assert_known_keys([name])

        try:
            return self.store[name]
        except KeyError:
            pass

        if name in self.dependency_graph.inputs:
            raise KeyError(f"Input has not been set: {name}")

        if name in self.dependency_graph.compute_nodes.keys():
            self.evaluate(targets=[name])

        return self.store[name]



import os
import sys
import typing as t
import boto3
from ..s3 import temp_file_on_s3

if t.TYPE_CHECKING:
    from mypy_boto3_s3.literals import RegionName
    from mypy_boto3_lambda.literals import RuntimeType
    from mypy_boto3_lambda.type_defs import FunctionCodeTypeDef


def default_runtime() -> "RuntimeType":
    import sys

    return t.cast(
        "RuntimeType", f"python{sys.version_info.major}.{sys.version_info.minor}"
    )


def needs_s3_upload(path_to_zipfile: str) -> bool:
    return os.path.getsize(path_to_zipfile) > 50 * 2**20


def _create_or_update(
    local_path_to_zipfile: t.Optional[str],
    s3_path_to_zipfile: t.Optional[str],
    message: str,
    create_or_update_fn: t.Callable[["FunctionCodeTypeDef"], None],
    verbose: bool = False,
    s3_code_bucket: t.Optional[str] = None,
    force_upload_to_s3_code_bucket: bool = False,
) -> None:
    """
    Create or update a lambda function with the given zipfile. If the zipfile is larger
    than 50 MB, you must specify an `s3_code_bucket`.
    """
    if bool(local_path_to_zipfile) == bool(s3_path_to_zipfile):
        raise ValueError(
            "Either `local_path_to_zipfile` or `s3_path_to_zipfile` must be provided (and not both)"
        )

    def pif(x: str) -> None:
        if verbose:
            print(x, file=sys.stderr)

    if s3_path_to_zipfile:
        if s3_code_bucket is None:
            raise ValueError(
                "When `s3_path_to_zipfile` is provided, `s3_code_bucket` is required"
            )
        create_or_update_fn({"S3Bucket": s3_code_bucket, "S3Key": s3_path_to_zipfile})
        pif(message)
        return

    if local_path_to_zipfile is None:
        raise ValueError("Shouldn't get here")

    if not os.path.isfile(local_path_to_zipfile):
        raise ValueError(f"Local zip file does not exist: {local_path_to_zipfile}")

    if needs_s3_upload(local_path_to_zipfile) or force_upload_to_s3_code_bucket:
        if s3_code_bucket is None:
            raise ValueError(
                """
                When zipfile is larger than 50 MB,
                or force_upload_to_s3_code_bucket is True,
                then s3_code_bucket is required
                """
            )
        with temp_file_on_s3(
            local_path=local_path_to_zipfile, bucket=s3_code_bucket, verbose=verbose
        ) as temp_key:
            create_or_update_fn({"S3Bucket": s3_code_bucket, "S3Key": temp_key})
            pif(message)
    else:
        with open(local_path_to_zipfile, "rb") as f:
            zipfile_contents = f.read()
        pif(f"Uploading {local_path_to_zipfile} to Lambda")
        create_or_update_fn({"ZipFile": zipfile_contents})
        pif(message)


def perform_create(
    aws_region: "RegionName",
    handler: str,
    function_name: str,
    role: str,
    local_path_to_zipfile: t.Optional[str] = None,
    s3_path_to_zipfile: t.Optional[str] = None,
    timeout: t.Optional[int] = None,
    memory_size: t.Optional[int] = None,
    runtime: t.Optional["RuntimeType"] = None,
    env_vars: dict[str, str] = {},
    s3_code_bucket: t.Optional[str] = None,
    verbose: bool = False,
    force_upload_to_s3_code_bucket: bool = False,
    wait_until_active: bool = True,
    wait_until_active_timeout_seconds: int = 30,
) -> None:
    if bool(local_path_to_zipfile) == bool(s3_path_to_zipfile):
        raise ValueError(
            "Either `local_path_to_zipfile` or `s3_path_to_zipfile` must be provided (and not both)"
        )

    client = boto3.client("lambda", region_name=aws_region)

    def create(code_arguments: "FunctionCodeTypeDef") -> None:
        extra_options = {}
        if timeout is not None:
            extra_options["Timeout"] = timeout
        if memory_size is not None:
            extra_options["MemorySize"] = memory_size

        client.create_function(
            FunctionName=function_name,
            Runtime=runtime or default_runtime(),
            Role=role,
            Handler=handler,
            Environment={"Variables": env_vars},
            Code=code_arguments,
            **extra_options,  # type: ignore[arg-type]
        )

    _create_or_update(
        local_path_to_zipfile=local_path_to_zipfile,
        s3_path_to_zipfile=s3_path_to_zipfile,
        message="Lambda function created",
        create_or_update_fn=create,
        verbose=verbose,
        s3_code_bucket=s3_code_bucket,
        force_upload_to_s3_code_bucket=force_upload_to_s3_code_bucket,
    )

    if wait_until_active:
        client.get_waiter("function_active").wait(
            FunctionName=function_name,
            WaiterConfig={
                "Delay": 1,
                "MaxAttempts": wait_until_active_timeout_seconds,
            },
        )


def perform_update_code(
    aws_region: "RegionName",
    function_name: str,
    local_path_to_zipfile: t.Optional[str] = None,
    s3_path_to_zipfile: t.Optional[str] = None,
    s3_code_bucket: t.Optional[str] = None,
    verbose: bool = False,
    force_upload_to_s3_code_bucket: bool = False,
    wait_until_updated: bool = True,
    wait_until_updated_timeout_seconds: int = 30,
) -> None:
    if bool(local_path_to_zipfile) == bool(s3_path_to_zipfile):
        raise ValueError(
            "Either `local_path_to_zipfile` or `s3_path_to_zipfile` must be provided (and not both)"
        )

    client = boto3.client("lambda", region_name=aws_region)

    def update(code_arguments: "FunctionCodeTypeDef") -> None:
        client.update_function_code(FunctionName=function_name, **code_arguments)

    _create_or_update(
        local_path_to_zipfile=local_path_to_zipfile,
        s3_path_to_zipfile=s3_path_to_zipfile,
        message="Lambda function code updated",
        create_or_update_fn=update,
        verbose=verbose,
        s3_code_bucket=s3_code_bucket,
        force_upload_to_s3_code_bucket=force_upload_to_s3_code_bucket,
    )

    if wait_until_updated:
        client.get_waiter("function_updated").wait(
            FunctionName=function_name,
            WaiterConfig={
                "Delay": 1,
                "MaxAttempts": wait_until_updated_timeout_seconds,
            },
        )



import typing as t
from abc import ABC, abstractmethod


class Destination(ABC):
    @abstractmethod
    def send(
        self, message_key: t.Any, output_message: t.Any
    ) -> None:  # pragma: no cover
        pass



from collections import namedtuple

InferredFunctionEnvironment = namedtuple(
    "InferredFunctionEnvironment",
    [
        "function_name",
        "version",
        "sha1",
    ],
)


def infer_function_environment_from_ci_environment() -> InferredFunctionEnvironment:
    import os
    import re
    from semver import Version

    # CircleCI is supported though contributions are welcome for other CI's.
    tag = os.environ["CIRCLE_TAG"]

    match = re.search(r"([a-z0-9-]+)@(.+)$", tag)
    if not match:
        raise ValueError(
            f'Malformed tag: {tag} (expected e.g. "my-lambda-function@1.2.3")'
        )
    function_name, version = match.groups()
    if not Version.is_valid(version):
        raise ValueError(f"Version in tag {version} is not valid semver")

    return InferredFunctionEnvironment(
        function_name=function_name,
        version=version,
        sha1=None,
    )


def infer_commit_from_ci_environment(function_name: str) -> InferredFunctionEnvironment:
    import os

    sha1 = os.environ["CIRCLE_SHA1"]

    return InferredFunctionEnvironment(
        function_name=function_name,
        version=None,
        sha1=sha1,
    )



def format_time(seconds: float) -> str:
    minutes, seconds = divmod(seconds, 60)
    if minutes > 0:
        return "{} min, {} sec".format(minutes, seconds)
    else:
        return "{} sec".format(seconds)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['werkit',
 'werkit.aws_lambda',
 'werkit.common',
 'werkit.compute',
 'werkit.compute.destination',
 'werkit.compute.graph',
 'werkit.orchestrator',
 'werkit.orchestrator.orchestrator_lambda',
 'werkit.orchestrator.testing',
 'werkit.orchestrator.testing.worker_service',
 'werkit.scripts']

package_data = \
{'': ['*'],
 'werkit.compute': ['generated/*'],
 'werkit.compute.graph': ['generated/*'],
 'werkit.orchestrator.orchestrator_lambda': ['generated/*'],
 'werkit.orchestrator.testing.worker_service': ['generated/*']}

install_requires = \
['jsonschema>=4.1.2,<5.0.0', 'missouri==0.3.0', 'semver==3.0.1']

extras_require = \
{'aws-lambda-build': ['executor>=21.0'],
 'cli': ['click>=8.0.3,<9.0.0'],
 'client': ['boto3==1.20.32'],
 'compute-graph': ['artifax==0.5'],
 'compute-graph:python_version < "3.10"': ['typing-extensions>=4'],
 'lambda-common': ['harrison>=2.0,<3.0'],
 'rds-graphile-worker': ['rds-graphile-worker-client>=0.1.1,<0.2.0']}

setup_kwargs = {
    'name': 'werkit',
    'version': '0.35.2',
    'description': 'Toolkit for encapsulating Python-based computation into deployable and distributable tasks',
    'long_description': 'None',
    'author': 'Paul Melnikow',
    'author_email': 'github@paulmelnikow.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/metabolize/werkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4',
}


setup(**setup_kwargs)



from __future__ import annotations
import inspect
import numbers
import typing as t


from ._built_in_type import (
    BuiltInValueType,
    BuiltInValueTypeName,
    built_in_value_type_with_name,
    coerce_value_to_builtin_type,
    is_built_in_value_type,
)
from ._custom_type import CustomType

if t.TYPE_CHECKING:  # pragma: no cover
    from jsonschema import Draft7Validator


def _attrs_of_type(obj: t.Any, _type: t.Type[AttrType]) -> dict[str, AttrType]:
    return {
        name: getattr(obj, name)
        for name in dir(obj)
        if not name.startswith("__") and isinstance(getattr(obj, name), _type)
    }


def _find_duplicates(items: list[str]) -> list[str]:
    from collections import Counter

    counter = Counter(items)
    return [item for item in counter if counter[item] > 1]


AnyValueType = t.Union[BuiltInValueType, t.Type[CustomType]]


class BaseNode:
    def __init__(self, value_type: AnyValueType):
        if is_built_in_value_type(value_type) or (
            inspect.isclass(value_type) and issubclass(value_type, CustomType)
        ):
            self.value_type = value_type
        else:
            raise ValueError(
                "Expected value type to be bool, int, float, str, or a subclass of CustomType"
            )

    @property
    def value_type_is_built_in(self) -> bool:
        return is_built_in_value_type(self.value_type)

    @property
    def value_type_name(self) -> str:
        if self.value_type_is_built_in:
            if self.value_type is bool:
                return "boolean"
            elif self.value_type in (int, float, numbers.Number):
                return "number"
            elif self.value_type is str:
                return "string"
            else:
                raise ValueError("How did we get here?")
        else:
            return t.cast(type[CustomType], self.value_type).name()

    def deserialize(self, value: t.Any) -> t.Any:
        if self.value_type_is_built_in:
            return value
        else:
            value_type = t.cast(t.Type[CustomType], self.value_type)
            value_type.validate(value)
            return value_type.deserialize(value)

    def normalize(self, name: str, value: t.Any) -> t.Any:
        if self.value_type_is_built_in:
            return coerce_value_to_builtin_type(
                name=name,
                value_type=t.cast(BuiltInValueType, self.value_type),
                value=value,
            )
        else:
            # TODO: Perhaps catch and re-throw to improve the error message.
            return t.cast(t.Type[CustomType], self.value_type).normalize(value)

    def serialize_value(self, value: t.Any) -> t.Any:
        if self.value_type_is_built_in:
            return value
        else:
            value_type = t.cast(t.Type[CustomType], self.value_type)
            serialized = value_type.serialize(value)
            value_type.validate(serialized)
            return serialized


class InputJSONType(t.TypedDict):
    valueType: str


class Input(BaseNode):
    def serialize(self) -> InputJSONType:
        return {"valueType": self.value_type_name}


class ComputeNodeJSONType(t.TypedDict):
    valueType: str
    dependencies: t.List[str]


class ComputeNode(BaseNode):
    # TODO: Use a narrower type for `method`.
    def __init__(self, method: t.Callable, value_type: AnyValueType):
        super().__init__(value_type=value_type)

        self.method = method
        self.dependencies = [x for x in inspect.signature(method).parameters.keys()][1:]

    def bind(self, instance: ComputeNode) -> t.Callable:
        import functools

        return functools.partial(self.method, instance)

    def serialize(self) -> ComputeNodeJSONType:
        return {
            "valueType": self.value_type_name,
            "dependencies": self.dependencies,
        }


class Intermediate(ComputeNode):
    pass


class Output(ComputeNode):
    pass


def intermediate(value_type: AnyValueType) -> t.Callable[[t.Callable], Intermediate]:
    def decorator(method: t.Callable[..., t.Any]) -> Intermediate:
        return Intermediate(method, value_type)

    return decorator


def output(value_type: AnyValueType) -> t.Callable[[t.Callable], Output]:
    def decorator(method: t.Callable[..., t.Any]) -> Output:
        return Output(method, value_type)

    return decorator


AttrType = t.TypeVar("AttrType")


class DependencyGraphJSONType(t.TypedDict):
    schemaVersion: t.Literal[1]
    inputs: dict[str, InputJSONType]
    intermediates: dict[str, ComputeNodeJSONType]
    outputs: dict[str, ComputeNodeJSONType]


def dependency_graph_validator() -> "Draft7Validator":
    import os
    from jsonschema import Draft7Validator, RefResolver
    from missouri import json

    schema = json.load(
        os.path.join(os.path.dirname(__file__), "dependency-graph.schema.json")
    )
    resolver = RefResolver.from_schema(schema)
    return Draft7Validator(
        {"$ref": "#/definitions/DependencyGraphWithAnyTypes"}, resolver=resolver
    )


def assert_valid_dependency_graph_data(data: t.Any) -> DependencyGraphJSONType:
    dependency_graph_validator().validate(data)
    return data


def not_implemented() -> t.NoReturn:
    raise NotImplementedError("Deserialized compute nodes are not implemented")


NodeType = t.TypeVar("NodeType", Input, Intermediate, Output)


def create_unimplemented_node(
    cls: type[NodeType],
    value_type_name: str,
    custom_types: dict[str, type[CustomType]],
) -> NodeType:
    value_type: AnyValueType
    try:
        value_type = custom_types[value_type_name]
    except KeyError:
        value_type = built_in_value_type_with_name(
            t.cast(BuiltInValueTypeName, value_type_name)
        )

    # TODO: Figure out why narrowing the typevar doesn't allow the correct
    # return type to be inferred.
    if cls is Input:
        return Input(value_type=value_type)  # type: ignore[return-value]
    elif cls is Intermediate:
        return Intermediate(method=not_implemented, value_type=value_type)  # type: ignore[return-value]
    else:
        return Output(method=not_implemented, value_type=value_type)  # type: ignore[return-value]


class DependencyGraph:
    def __init__(
        self,
        inputs: dict[str, Input],
        intermediates: dict[str, Intermediate],
        outputs: dict[str, Output],
    ):
        self.inputs = inputs
        self.intermediates = intermediates
        self.outputs = outputs
        self.compute_nodes = dict(**self.intermediates, **self.outputs)
        self.all_nodes = dict(**self.inputs, **self.compute_nodes)

    @classmethod
    def from_class(cls: type[DependencyGraph], in_class: t.Type) -> "DependencyGraph":
        assert inspect.isclass(in_class)

        return cls(
            inputs=_attrs_of_type(in_class, Input),
            intermediates=_attrs_of_type(in_class, Intermediate),
            outputs=_attrs_of_type(in_class, Output),
        )

    @classmethod
    def deserialize(
        cls: type[DependencyGraph],
        data: DependencyGraphJSONType,
        custom_types: list[type[CustomType]] = [],
    ) -> "DependencyGraph":
        assert_valid_dependency_graph_data(data)

        custom_type_names = [item.name() for item in custom_types]
        duplicates = _find_duplicates(custom_type_names)
        if duplicates:
            raise ValueError(
                f"Duplicate custom type names found: {', '.join(duplicates)}"
            )
        keyed_custom_types = {item.name(): item for item in custom_types}

        return cls(
            inputs={
                name: create_unimplemented_node(
                    cls=Input,
                    value_type_name=input_data["valueType"],
                    custom_types=keyed_custom_types,
                )
                for name, input_data in data["inputs"].items()
            },
            intermediates={
                name: create_unimplemented_node(
                    cls=Intermediate,
                    value_type_name=input_data["valueType"],
                    custom_types=keyed_custom_types,
                )
                for name, input_data in data["intermediates"].items()
            },
            outputs={
                name: create_unimplemented_node(
                    cls=Output,
                    value_type_name=input_data["valueType"],
                    custom_types=keyed_custom_types,
                )
                for name, input_data in data["outputs"].items()
            },
        )

    def keys(self) -> t.List[str]:
        return list(self.inputs.keys()) + list(self.compute_nodes.keys())

    def serialize(self) -> DependencyGraphJSONType:
        return {
            "schemaVersion": 1,
            "inputs": {k: v.serialize() for k, v in self.inputs.items()},
            "intermediates": {k: v.serialize() for k, v in self.intermediates.items()},
            "outputs": {k: v.serialize() for k, v in self.outputs.items()},
        }



import typing as t

from ._dependency_graph import ComputeNode, Input
from ._state_manager import StateManager


# This decorator is hard to type, so we provide a protcol which can be used with
# `t.cast()`.

# https://github.com/python/typing/issues/213
# https://stackoverflow.com/a/75101725/893113
# https://stackoverflow.com/a/74041065/893113
# https://github.com/python/mypy/issues/3135


class DefaultStateManagerProtocol(t.Protocol):
    @property
    def state_manager(self) -> StateManager:  # pragma: no cover
        ...


T = t.TypeVar("T")


def bind_state_manager(
    attr_name: str = "state_manager",
) -> t.Callable[[type[T]], type[T]]:
    def decorator(cls):  # type: ignore[no-untyped-def]
        class Bound(cls):  # type: ignore[no-untyped-def]
            def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
                super().__init__(*args, **kwargs)
                setattr(self, attr_name, StateManager(self))

            def __getattribute__(self, name: str) -> t.Any:
                attr = object.__getattribute__(self, name)
                if isinstance(attr, Input) or isinstance(attr, ComputeNode):
                    state_manager = object.__getattribute__(self, attr_name)
                    return state_manager.get(name)
                else:
                    return attr

        return Bound

    return decorator
